/*
 * Copyright (c) 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Lawrence Berkeley Laboratory,
 * Berkeley, CA.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
static char rcsid[] =
"@(#) $Header: ed1.c,v 1.2 92/11/30 10:55:23 floyd Exp $ (LBL)";
#endif

#include <stdio.h>
#include <sys/types.h>

#include "real.h"
#include "packet.h"
#include "at.h"
#include "config.h"
#include "drop.h"
#include "pq.h"

extern double exp2();
extern double log2();

/*
 * Early drop parameters.
 */
struct edp {
	/*
	 * User supplied.
	 */
	int mean_pktsize;

	float th_min;
	float q_w;
	/*
	 * Computed as a function of user supplied paramters.
	 */
	float k;
	float ptc;		/* packet time constant */
};

/*
 * Early drop variables.
 */
struct edv {
	float v_th;
	float v_ave;		/* average queue size */
	float v_avelast;	/* used in computing average queue size */
	float v_slope;		/* used in computing average queue size */
	float v_r;			
	struct dlist *drops;
};

struct edstate {
	/*
	 * Static state.
	 */
	int maxq;
	int avg_pktsize;
	dropfn drop;
	char *plot_fname;
	FILE *plotfile;
	struct edp edp;
	int doubleq;
	int dqthresh;
	/*
	 * Dynamic state.
	 */
	int idle;
	double idletime;
	struct pq q;
	struct edv edv;
};
#define STATE(qd) ((struct edstate *)((qd)->qd_state))

static struct at_options options[] = {
	{ "queue-size", AT_INT, offsetof(struct edstate, maxq) },
	{ "plot-file", AT_NSTRING, offsetof(struct edstate, plot_fname) },
	{ "thresh", AT_FLOAT, offsetof(struct edstate, edp.th_min) },
	{ "mean_pktsize", AT_INT, offsetof(struct edstate, edp.mean_pktsize) },
	{ "q_weight", AT_FLOAT, offsetof(struct edstate, edp.q_w) },
	{ "dropmech", AT_DROP, offsetof(struct edstate, drop) },
	{ "doubleq", AT_BOOL, offsetof(struct edstate, doubleq) },
	{ "dqthresh", AT_INT, offsetof(struct edstate, dqthresh) },
	{ 0 },
};

static void
ed1_comp(stp, nqueued, m)
	struct edstate *stp;
	int nqueued;
	int m;
{
	float f, f_sl, f_old;

	f = stp->edv.v_ave;
	f_sl = stp->edv.v_slope;
/*  EWMA  */
        while (--m >= 1) {
                f_old = f;
                f *= (1.0 - stp->edp.q_w);
        }
        f_old = f;
        f *= (1.0 - stp->edp.q_w);
        f += stp->edp.q_w * (float)nqueued;
	stp->edv.v_ave = f;
	stp->edv.v_slope = f_sl;
}

void
ed1_init(struct qdisc *qd, double ifspeed)
{
	struct edstate *stp = (struct edstate *)malloc(sizeof(*stp));

	bzero((char *)stp, sizeof(*stp));
	atl_init(qd->qd_nb, qd->qd_attr, options, (char *)stp);
	qd->qd_state = (void *)stp;

	/* Packet time constant -- units are pkts/usec */
	/* 0 when average queue first exceeds thresh */
	stp->edp.ptc = 1 / ifspeed * stp->edp.mean_pktsize;

	stp->edv.v_th = stp->edp.th_min;
	stp->edv.v_ave = 0.0;
	stp->edv.v_slope = 0.0;
	stp->edv.drops = 0;
	stp->edv.v_avelast = 0.0;
	pqinit(&stp->q, stp->plot_fname, qd);
	stp->plotfile = stp->q.file;
}

static void
ed1_plot(stp)
	struct edstate *stp;
{
	float t = realtime();

	fprintf(stp->plotfile, "a %g %g\n", t, stp->edv.v_ave);
	fprintf(stp->plotfile, "s %g %g\n", t, stp->edv.v_slope);
}

struct packet *
ed1_send(struct qdisc *qd)
{
	struct edstate *stp = STATE(qd);
	struct packet *pkt;

	if (stp->doubleq)
		pkt = pqthdeq(&STATE(qd)->q, stp->dqthresh);
	else
		pkt = pqdeq(&STATE(qd)->q);

	if (pkt == 0) {
		stp->idle = 1;
		stp->idletime = realtime();
	} else
		stp->idle = 0;
	return pkt;
}

void
ed1_recv(struct qdisc *qd, struct packet *pkt)
{
	struct edstate *stp = STATE(qd);
	struct pq *q = &stp->q;
	int m;

        if (stp->idle) {
                stp->idle = 0;
                m = stp->edp.ptc * (realtime() - stp->idletime);
        } else
                m = 0;

	ed1_comp(stp, q->len, m + 1);

	if (stp->edv.v_ave >= stp->edp.th_min) { 
		/*
		 * Drop each red packet if ave_queue_size too high.
		 */
	  	if (pkt != 0 && pkt->pri==2) {
		    qd_drop(qd, pkt);
		    pkt = 0;
	  	}
	}
	/*
	 * If we didn't drop the packet above, send it to the interface,
	 * checking for absolute queue overflow.
	 */
	if (pkt != 0) {
		pqenq(q, pkt);
		if (q->len > stp->maxq) {
			qd_drop(qd, pqrm(q, (*stp->drop)(q->len)));
		}
	}
	if (stp->plotfile)
		ed1_plot(stp);
}

int
ed1_len(struct qdisc *qd)
{
	return STATE(qd)->q.len;
}
