/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by the Network Research
 * 	Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: flow.cc,v 1.17 96/11/06 14:42:13 floyd Exp $ (LBL)";
#endif

#include <stdlib.h>
#include "flow.h"

// tcl linkage
class FlowManagerMatcher : public Matcher {
    public:
        inline FlowManagerMatcher() : Matcher("flowmgr") { }
        TclObject* match(const char* id) {
		if (id == 0) {
			printf("none specified\n");
			return 0;
		}
                if (strcmp(id, "class") == 0)
                        return (new ClassFlowManager);
                else if (strcmp(id, "src") == 0)
                        return (new SrcFlowManager);
                else if (strcmp(id, "dst") == 0)
                        return (new DstFlowManager);
                else if (strcmp(id, "srcdst") == 0)
                        return (new SrcDstFlowManager);
                else if (strcmp(id, "dstsrc") == 0)
                        return (new SrcDstFlowManager);
                else if (strcmp(id, "srcdstcls") == 0)
                        return (new SrcDstClassFlowManager);
                else if (strcmp(id, "redsrcdstcls") == 0)
                        return (new SrcDstClassFlowManager(FTYPE_RED));
                return (0);
        }
} flowmgrmatcher;

struct drpinfo Flow::droptotals_[MAXCATEGORY];
struct drpinfo Flow::alldrops_;
struct arrinfo Flow::allarrivals_;
u_int Flow::ndcat_;

/*
 * tcl "member functions" on a flowmgr object
 */
int FlowManager::command(int argc, const char*const* argv)
{
        Tcl& tcl = Tcl::instance();
        if (argc == 2) {
                if (strcmp(argv[1], "detach") == 0) {
                        channel_ = 0;
                        return (TCL_OK);
                }
                if (strcmp(argv[1], "flush") == 0) {
                        Tcl_Flush(channel_);
                        return (TCL_OK);
                }
                if (strcmp(argv[1], "resetcounters") == 0) {
                        resetcounters();
                        return (TCL_OK);
                }
		// usage: dump
		//	dumps all categories... also see "dump categ" below
                if (strcmp(argv[1], "dump") == 0) {
                        dumpflows();
			return (TCL_OK);
                }
        } else if (argc == 3) {
		// attach a file for output of dump() operation
                if (strcmp(argv[1], "attach") == 0) {
                        int mode;
                        const char* id = argv[2];
                        channel_ = Tcl_GetChannel(tcl.interp(), (char*)id,
                                                  &mode);
                        if (channel_ == 0) {
                                tcl.resultf("flowmgr: can't attach %s for writing"
, id);
                                return (TCL_ERROR);
                        }
                        return (TCL_OK);
                }
		// define the number of drop categories
		if (strcmp(argv[1], "ndropcats") == 0) {
			int ncats = atoi(argv[2]);
			if (ncats <= 0 || ncats > MAXCATEGORY) {
				tcl.resultf("flowmgr: ndrop cats %d invalid [max %d]\n",
					ncats, MAXCATEGORY);
				return (TCL_ERROR);
			}
			Flow::ndcat_ = ncats;
			return (TCL_OK);
		}

		// tot pkt drops theshold above which to invoke pcallback
		if (strcmp(argv[1], "total-drops-pthresh") == 0) {
			int thresh = atoi(argv[2]);
			if (thresh < 0) {
				tcl.resultf("flowmgr: invalid drop-thresh: %d",
					thresh);
				return (TCL_ERROR);
			}
			Flow::alldrops_.pkt_thresh_ = thresh;
			return (TCL_OK);
		}
		// tot bytes in drops thresh above which to invoke bcallback
		if (strcmp(argv[1], "total-drops-bthresh") == 0) {
			int thresh = atoi(argv[2]);
			if (thresh < 0) {
				tcl.resultf("flowmgr: invalid drop-thresh: %d",
					thresh);
				return (TCL_ERROR);
			}
			Flow::alldrops_.byte_thresh_ = thresh;
			return (TCL_OK);
		}
		// define the total packet drops handler
                if (strcmp(argv[1], "total-drops-pcallback") == 0) {
                        const char* cb = argv[2];
                        delete Flow::alldrops_.pcallback_;
                        if (*cb == 0)
                                Flow::alldrops_.pcallback_ = 0;
                        else {
                                Flow::alldrops_.pcallback_ = new char[strlen(cb) + 1];
                                strcpy(Flow::alldrops_.pcallback_, cb);
                        }
                        return (TCL_OK);
                }
		// define the total bytes in packet drops handler
                if (strcmp(argv[1], "total-drops-bcallback") == 0) {
                        const char* cb = argv[2];
                        delete Flow::alldrops_.bcallback_;
                        if (*cb == 0)
                                Flow::alldrops_.bcallback_ = 0;
                        else {
                                Flow::alldrops_.bcallback_ = new char[strlen(cb) + 1];
                                strcpy(Flow::alldrops_.bcallback_, cb);
                        }
                        return (TCL_OK);
                }


		// total pkt arrivals thresh above which to invoke pcallback
		if (strcmp(argv[1], "total-arrivals-pthresh") == 0) {
			int thresh = atoi(argv[2]);
			if (thresh < 0) {
				tcl.resultf("flowmgr: invalid arr-thresh: %d",
					thresh);
				return (TCL_ERROR);
			}
			Flow::allarrivals_.pkt_thresh_ = thresh;
			return (TCL_OK);
		}
		// total bytes in arrivals above which to invoke bcallback
		if (strcmp(argv[1], "total-arrivals-bthresh") == 0) {
			int thresh = atoi(argv[2]);
			if (thresh < 0) {
				tcl.resultf("flowmgr: invalid arr-thresh: %d",
					thresh);
				return (TCL_ERROR);
			}
			Flow::allarrivals_.byte_thresh_ = thresh;
			return (TCL_OK);
		}
		// define the total pkt arrivals handler
                if (strcmp(argv[1], "total-arrivals-pcallback") == 0) {
                        const char* cb = argv[2];
                        delete Flow::allarrivals_.pcallback_;
                        if (*cb == 0)
                                Flow::allarrivals_.pcallback_ = 0;
                        else {
                                Flow::allarrivals_.pcallback_ = new char[strlen(cb) + 1];
                                strcpy(Flow::allarrivals_.pcallback_, cb);
                        }
                        return (TCL_OK);
                }
		// define the total bytes in pkt arrivals handler
                if (strcmp(argv[1], "total-arrivals-bcallback") == 0) {
                        const char* cb = argv[2];
                        delete Flow::allarrivals_.bcallback_;
                        if (*cb == 0)
                                Flow::allarrivals_.bcallback_ = 0;
                        else {
                                Flow::allarrivals_.bcallback_ = new char[strlen(cb) + 1];
                                strcpy(Flow::allarrivals_.bcallback_, cb);
                        }
                        return (TCL_OK);
                }
		// usage: dump cat#
                if (strcmp(argv[1], "dump") == 0) {
			int categ = atoi(argv[2]);
			if (categ < 0 || categ > MAXCATEGORY) {
				tcl.resultf("flowmgr: invalid categ: %d",
					categ);
				return (TCL_ERROR);
			}
                        dumpflows(categ);
			return (TCL_OK);
                }

/////
	} else if (argc == 4) {
		if (strcmp(argv[1], "cat-drops-pthresh") == 0) {
			int category = atoi(argv[2]);
			int thresh = atoi(argv[3]);
			if (thresh < 0 || category > MAXCATEGORY) {
				tcl.resultf("flowmgr: invalid arr-thresh: %d",
					thresh);
				return (TCL_ERROR);
			}
			Flow::droptotals_[category].pkt_thresh_ = thresh;
			return (TCL_OK);
		}
		if (strcmp(argv[1], "cat-drops-bthresh") == 0) {
			int category = atoi(argv[2]);
			int thresh = atoi(argv[3]);
			if (thresh < 0 || category > MAXCATEGORY) {
				tcl.resultf("flowmgr: invalid arr-thresh: %d",
					thresh);
				return (TCL_ERROR);
			}
			Flow::droptotals_[category].byte_thresh_ = thresh;
			return (TCL_OK);
		}
                if (strcmp(argv[1], "cat-drops-pcallback") == 0) {
			int category = atoi(argv[2]);
                        const char* cb = argv[3];
			if (category < 0 || category > MAXCATEGORY) {
				tcl.resultf("flowmgr: invalid categ: %d",
					category);
				return (TCL_ERROR);
			}
                        delete Flow::droptotals_[category].pcallback_;
                        if (*cb == 0)
                                Flow::droptotals_[category].pcallback_ = 0;
                        else {
                                Flow::droptotals_[category].pcallback_ = new char[strlen(cb) + 1];
                                strcpy(Flow::droptotals_[category].pcallback_, cb);
                        }
                        return (TCL_OK);
                }
                if (strcmp(argv[1], "cat-drops-bcallback") == 0) {
			int category = atoi(argv[2]);
                        const char* cb = argv[3];
			if (category < 0 || category > MAXCATEGORY) {
				tcl.resultf("flowmgr: invalid categ: %d",
					category);
				return (TCL_ERROR);
			}
                        delete Flow::droptotals_[category].bcallback_;
                        if (*cb == 0)
                                Flow::droptotals_[category].bcallback_ = 0;
                        else {
                                Flow::droptotals_[category].bcallback_ = new char[strlen(cb) + 1];
                                strcpy(Flow::droptotals_[category].bcallback_, cb);
                        }
                        return (TCL_OK);
                }
	}
	tcl.resultf("ns flowmgr object: no such command %s", argv[1]);
	return (TCL_ERROR);
}


// ctors and dtors
FlowManager::FlowManager(u_int ftype = 0) : nflows_(0), channel_(0),
	ftype_(ftype)
{
	memset(flows_, '\0', sizeof(flows_));
}

FlowManager::~FlowManager()
{
	register i;
	for (i = 0; i < nflows_; i++)
		delete flows_[i];
}

void
FlowManager::resetcounters()
{
	register Flow **p;
	register int i;
	for (i = 0, p = flows_; i < nflows_; i++, p++)
		(*p)->resetcounters();
	return;
}

// dump a particular category's flows

int
FlowManager::dumpflows(int categ)
{
	register Flow **p;
	register int i, rval = TCL_OK;

	for (i = 0, p = flows_; i < nflows_; i++, p++)
		if ((*p)->dump(channel_, i, categ) != TCL_OK)
			rval = TCL_ERROR;

	return (rval);
}

// dump all flows + categories

int
FlowManager::dumpflows()
{
	register int i, rval = TCL_OK;
	int ndcat;

	if (*flows_) {
		ndcat = flows_[0]->ndcat();
		for (i = 0; i < ndcat; i++)
			if (dumpflows(i) != TCL_OK)
				rval = TCL_ERROR;
	}

	return (rval);
}

Flow*
FlowManager::findflow(Packet* pkt)
{
	register Flow **p;
	register int i;

	for (i = 0, p = flows_; i < nflows_; i++, p++)
		if (match(pkt, *p))
			return(*p);

	return (0);
}

Flow *
FlowManager::newflow(Packet* pkt)
{
	if (++nflows_ == MAXFLOWS) {
		fprintf(stderr, "exceeded max number of flows (%d)\n", MAXFLOWS);
		--nflows_;
		return NULL;
	}
	Flow *f;
/* XXX */
if (ftype_ == 0)
	f = flows_[nflows_-1] = new Flow(this);
else if (ftype_ == FTYPE_RED)
	f = flows_[nflows_-1] = new RedFlow(this);
else {
fprintf(stderr, "unknown ftype_ %d\n", ftype_);
}
	if (f == NULL)
		return NULL;
	f->type_ = pkt->type_;
	f->class_ = pkt->class_;
	f->src_ = pkt->src_;
	f->dst_ = pkt->dst_;

	return f;
}

inline void
FlowManager::clearflows()
{
	register i;

	for (i = 0; i < nflows_; i++)
		delete flows_[i];
	nflows_ = 0;
	return;
}

/************************* flow class ****************************/





// log stats associated with a packet arrival on this flow
void
Flow::arrival(Packet* pkt)
{
	//
	// stats local to this flow
	//

	arrivalinfo_.pkt_arrivals_++;
	arrivalinfo_.byte_arrivals_ += pkt->size_;

	int pkts = arrivalinfo_.pkt_arrivals_;
	int bytes = arrivalinfo_.byte_arrivals_;
	int pthresh = arrivalinfo_.pkt_thresh_;
	int bthresh = arrivalinfo_.byte_thresh_;

	if (pkts >= pthresh && pthresh > 0 && arrivalinfo_.pcallback_)
		docallback(arrivalinfo_.pcallback_, NOCATEGORY);

	if (bytes >= bthresh && bthresh > 0 && arrivalinfo_.bcallback_)
		docallback(arrivalinfo_.bcallback_, NOCATEGORY);

	//
	// stats across all flows
	//

	allarrivals_.pkt_arrivals_++;
	allarrivals_.byte_arrivals_ += pkt->size_;

	pkts = allarrivals_.pkt_arrivals_;
	bytes = allarrivals_.byte_arrivals_;
	pthresh = allarrivals_.pkt_thresh_;
	bthresh = allarrivals_.byte_thresh_;

	if (pkts >= pthresh && pthresh > 0 && allarrivals_.pcallback_)
		docallback(allarrivals_.pcallback_, NOCATEGORY);

	if (bytes >= bthresh && bthresh > 0 && allarrivals_.bcallback_)
		docallback(allarrivals_.bcallback_, NOCATEGORY);

	return;
}



// drop a packet of a particular category, keep a number of stats
// and if they exceed any specified thresholds, call a tcl handler

void
Flow::drop(Packet* pkt, int category = 0)
{
	if (category > ndcat_ || category > MAXCATEGORY || category < 0) {
		fprintf(stderr, "drop category %d invalid-- max [%d]\n",
			category, MAXCATEGORY);
		return;
	}

	// per-flow stats
	dropinfo_[category].pkt_drops_++;
	dropinfo_[category].byte_drops_ += pkt->size_;

	int pkts = dropinfo_[category].pkt_drops_;
	int bytes = dropinfo_[category].byte_drops_;
	int pthresh = dropinfo_[category].pkt_thresh_;
	int bthresh = dropinfo_[category].byte_thresh_;

	if (pkts >= pthresh && pthresh > 0 && dropinfo_[category].pcallback_)
		docallback(dropinfo_[category].pcallback_, category);

	if (bytes >= bthresh && bthresh > 0 && dropinfo_[category].bcallback_)
		docallback(dropinfo_[category].bcallback_, category);


	// per-category stats
	droptotals_[category].pkt_drops_++;
	droptotals_[category].byte_drops_ += pkt->size_;

	pkts = droptotals_[category].pkt_drops_;
	bytes = droptotals_[category].byte_drops_;
	pthresh = droptotals_[category].pkt_thresh_;
	bthresh = droptotals_[category].byte_thresh_;

	if (pkts >= pthresh && pthresh > 0 && droptotals_[category].pcallback_)
		docallback(droptotals_[category].pcallback_, category);

	if (bytes >= bthresh && bthresh > 0 && droptotals_[category].bcallback_)
		docallback(droptotals_[category].bcallback_, category);

	// all stats
	alldrops_.pkt_drops_++;
	alldrops_.byte_drops_ += pkt->size_;

	pkts = alldrops_.pkt_drops_;
	bytes = alldrops_.byte_drops_;
	pthresh = alldrops_.pkt_thresh_;
	bthresh = alldrops_.byte_thresh_;

	if (pkts >= pthresh && pthresh > 0 && alldrops_.pcallback_)
		docallback(alldrops_.pcallback_, NOCATEGORY);

	if (bytes >= bthresh && bthresh > 0 && alldrops_.bcallback_)
		docallback(alldrops_.bcallback_, NOCATEGORY);

	return;
}

int
Flow::dump(Tcl_Channel tc, int fid, int categ)
{
	int n, k;
	if (tc == 0)
		return TCL_ERROR;

	n = formatc(fid, categ);

	wrk_[n] = '\n';
	wrk_[n+1] = '\0';
	(void)Tcl_Write(tc, wrk_, n+1);
	wrk_[n] = '\0';

	return (TCL_OK);
}

int
Flow::formatc(int fid, int category)
{
/* #define DEBUGFLOW */
#ifdef DEBUGFLOW
	printf("Flow::format: [fid %d, cat %d]: parr: %d, barr: %d, pdrop: %d, bdrop: %d, total_parrivals:%d, total_barrivals: %d\n",
		fid, category,
		arrivalinfo_.pkt_arrivals_,
		arrivalinfo_.byte_arrivals_,
		dropinfo_[category].pkt_drops_,
		dropinfo_[category].byte_drops_,
		allarrivals_.pkt_arrivals_,
		allarrivals_.byte_arrivals_);
	fflush(stdout);

#endif
	// format [17 values]:
	//	time, flow-id, category, type, class, src, dst,
	//	arrivals-this-flow(pkts), arrivals-this-flow(bytes),
	//	drops-this-flow-and-cat(pkts), drops-this-flow-and-cat(bytes),
	//	all-arrivals(pkts), all-arrivals(bytes),
	//	all-drops-this-cat(pkts), all-drops-this-cat(bytes)
	//	all-drops(pkts), all-drops(bytes)

	sprintf(wrk_, "%8.3f %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d",
			Scheduler::instance().clock(), fid, category,
			type_, class_, src_, dst_,
			arrivalinfo_.pkt_arrivals_,
			arrivalinfo_.byte_arrivals_,
			dropinfo_[category].pkt_drops_,
			dropinfo_[category].byte_drops_,
			allarrivals_.pkt_arrivals_,
			allarrivals_.byte_arrivals_,
			droptotals_[category].pkt_drops_,
			droptotals_[category].byte_drops_,
			alldrops_.pkt_drops_,
			alldrops_.byte_drops_);
	return(strlen(wrk_));
}

void
Flow::docallback(char *callback, int categ)
{
	if (callback != 0) {
		Tcl& tcl = Tcl::instance();
		if (categ == NOCATEGORY)
		    tcl.evalf("%s %s",
			callback, fm_->TclObject::name());
		else
		    tcl.evalf("%s %s %d",
			callback, fm_->TclObject::name(), categ);
	} else {
		fprintf(stderr, "docallback(): null callback\n");
	}
	return;
}

/*******************************************************************/

void
RedFlow::arrival(Packet* pkt)
{
	// stats local to this flow
	arrivalinfo_.byte_arrivals_ += pkt->size_;
	// stats across all flows
	allarrivals_.byte_arrivals_ += pkt->size_;
	return;
}

void
RedFlow::drop(Packet* pkt, int category = 0) {
	if (category > ndcat_ || category > MAXCATEGORY || category < 0) {
		fprintf(stderr, "drop category %d invalid-- max [%d]\n",
			category, MAXCATEGORY);
		return;
	}

	// per-flow stats
        if (category == 0) {
		dropinfo_[category].byte_drops_ += pkt->size_;
		droptotals_[category].byte_drops_ += pkt->size_;
        } else if (category == 1) {
		dropinfo_[category].pkt_drops_++;
        }
	// per-category stats
	droptotals_[category].pkt_drops_++;
	// all stats
	alldrops_.pkt_drops_++;

	int pkts = alldrops_.pkt_drops_;
	int pthresh = alldrops_.pkt_thresh_;

	if (pkts >= pthresh && pthresh > 0 && alldrops_.pcallback_)
		// Notes on work in progress:
		// Here I need to call a procedure to munch on the
		// information.  But the procedure needs to be
		// in red.cc, because I need information about 
		// the average queue size, as a measure of the level
		// of congestion.  I need to pass the information
		// from dump() to red.cc.  
		// First, the fraction of drops that were forced.
		// Second, for each flowID with drops, the single
		// measure of bandwidth for that flowID:
		//  (dropinfo_[0].byte_drops_ / droptotals_[0].byte_drops_) 
		//    * droptotals_[0].pkt_drops_ / alldrops_.pkt_drops_
		// + (dropinfo_[1].pkt_drops_ / droptotals_[1].pkt_drops_) 
		//  * droptotals_[1].pkt_drops_ / alldrops_.pkt_drops_
		docallback(alldrops_.pcallback_, NOCATEGORY);

	return;
}

