/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by the Network Research
 * 	Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

//
// FlowManager classes and a Flow class
//	idea is that a routing entity may create a flow manager
//	and pass packets to it.  The manager will keep track of
//	various statistics regarding the flow
//
// This is presently experimental
//

#ifndef ns_flow_h
#define ns_flow_h
#define REDDROPS		/* forced or unforced packet drops */

#include <stdio.h>
#include "object.h"
#include "packet.h"

class Tcl;

//
// a flow manager-- one of these per node interested in monitoring flows
//
// for now, the flows are maintained as a simple array-- if the
// number of flows is actually fairly large, this should become a hash
// table or other structure capable of faster retrieval
//
#define	MAXFLOWS	1024
#define	MAXCATEGORY	3
#define	NOCATEGORY	(-1)

class Flow;

struct drpinfo {
	u_int	pkt_drops_;
	u_int	byte_drops_;
	u_int	pkt_thresh_;
	u_int	byte_thresh_;
	char*	pcallback_;
	char*	bcallback_;
};

struct arrinfo {
	u_int	pkt_arrivals_;
	u_int	byte_arrivals_;
	u_int	pkt_thresh_;
	u_int	byte_thresh_;
	char*	pcallback_;
	char*	bcallback_;
};

#define	FTYPE_RED	1

class FlowManager : public NsObject {
    protected:
	u_int nflows_;		// # of current flows
	u_int ftype_;
	Tcl_Channel channel_;
	Flow* flows_[MAXFLOWS];
    public:
	FlowManager(u_int ftype = 0);
	~FlowManager();
	int command(int argc, const char*const* argv);
	Flow* findflow(Packet*);
	virtual int match(Packet*, Flow*) = 0;
	Flow* newflow(Packet*);
	void resetcounters();
	void clearflows();
	int dumpflows(int);
	int dumpflows();
};


//
// flow structure-- fields storing which corresponding packet header field
//	values must match to be on the same "flow".  the flow stucture is
//	used across all possible sorts of flows, where only the match()
//	function in the flow managers (above) is different
//	Also, this structure holds per-flow accumulated stats

class Flow {
    friend void FlowManager::newflow(Packet*);
    friend void FlowManager::command(int, const char*const*);
    private:
	int formatc(int, int);
	inline void zerocounters(struct drpinfo *p) {
		p->pkt_drops_ = p->byte_drops_ = 0;
	}
	inline void zerocounters(struct arrinfo *p) {
		p->pkt_arrivals_ = p->byte_arrivals_ = 0;
	}
    protected:
	u_int8_t	type_;
	u_int8_t	class_;
	u_short		src_;
	u_short		dst_;

	FlowManager*	fm_;

	// shared across all flow objects:
	static u_int 		ndcat_;				// # drop categ
	static struct drpinfo	droptotals_[MAXCATEGORY];	// per-cat
	static struct drpinfo	alldrops_;			// grand total
	static struct arrinfo	allarrivals_;			// grand total

	// per-flow stats
	struct drpinfo		dropinfo_[MAXCATEGORY];		// per-at
	struct arrinfo		arrivalinfo_;			// arrivals

	char			wrk_[256];

    public:
	Flow(FlowManager *fm) : fm_(fm) {
		memset(dropinfo_, '\0', sizeof(dropinfo_));
		memset(&arrivalinfo_, '\0', sizeof(dropinfo_));
		wrk_[0] = 0;
	}
	// docallback moved from private to public
	void docallback(char*, int);
	inline void resetcounters() {
		register i;

		for (i = 0; i < ndcat_; i++) {
			zerocounters(dropinfo_ + i);
			zerocounters(droptotals_ + i);
		}
		zerocounters(&arrivalinfo_);
		zerocounters(&allarrivals_);
		zerocounters(&alldrops_);
	}
	
	u_short src() { return (src_); }
	u_short dst() { return (dst_); }
	u_int8_t type() { return (type_); }
	u_int8_t fclass() { return (class_); }
		// can't be class(), a reserved word in C++

	virtual void drop(Packet* pkt, int category = 0);
	virtual void arrival(Packet* pkt);

	inline int pdrops(int cat = 0) { return(dropinfo_[cat].pkt_drops_); }
	inline int bdrops(int cat = 0) { return(dropinfo_[cat].byte_drops_); }
	inline int parrivals() { return(arrivalinfo_.pkt_arrivals_); }
	inline int barrivals() { return(arrivalinfo_.byte_arrivals_); }

	inline int total_pdrops(int ct) { return(droptotals_[ct].pkt_drops_); }
	inline int total_bdrops(int ct) { return(droptotals_[ct].byte_drops_); }
	inline int total_parrivals() { return(allarrivals_.pkt_arrivals_); }
	inline int total_barrivals() { return(allarrivals_.byte_arrivals_); }
	inline int ndcat() { return (ndcat_); }
	int dump(Tcl_Channel, int, int);
};

class RedFlow : public Flow {
    public:
	inline RedFlow(FlowManager *fm) : Flow(fm) { }
	void drop(Packet* pkt, int category = 0);
	void arrival(Packet* pkt);
};

class ClassFlowManager : public FlowManager {
    protected:
	inline int match(Packet* pkt, Flow *fp) {
		return (pkt->class_ == fp->fclass());
	};
};

class SrcFlowManager : public FlowManager {
    protected:
	inline int match(Packet* pkt, Flow *fp) {
		return (pkt->src_ == fp->src());
	};
};

class DstFlowManager : public FlowManager {
    protected:
	inline int match(Packet* pkt, Flow *fp) {
		return (pkt->dst_ == fp->dst());
	};
};

class SrcDstFlowManager : public FlowManager {
    protected:
	inline int match(Packet* pkt, Flow *fp) {
		return (pkt->src_ == fp->src() &&
			pkt->dst_ == fp->dst());
	};
};

class SrcDstClassFlowManager : public FlowManager {
    public:
	inline SrcDstClassFlowManager(u_int ftype = 0) : FlowManager(ftype) { }
    protected:
	inline int match(Packet* pkt, Flow *fp) {
		return (pkt->src_ == fp->src() &&
			pkt->dst_ == fp->dst() &&
			pkt->class_ == fp->fclass());
	};
};

#endif
