/*
 * Copyright (c) 1991-1994 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and the Network Research Group at
 *	Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Routing code for general topologies based on min-cost routing algorithm in
 * Bertsekas' book.  Written originally by S. Keshav, 7/18/88
 * (his work covered by identical UC Copyright)
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: route.cc,v 1.12 96/01/13 16:05:02 mccanne Exp $ (LBL)";
#endif

#include <stdlib.h>
#include "Tcl.h"
#include "link.h"
#include "node.h"

#define INFINITY 0x3fff

class RouteLogic : public TclObject {
public:
	RouteLogic() : nodes_(0) {}
	int command(int argc, const char*const* argv);
protected:
	void compute_routes();
	Node* nodes_;
};

class RouteLogicMatcher : public Matcher {
public:
	RouteLogicMatcher() : Matcher("route-logic") {}
	TclObject* match(const char* id) {
		return (new RouteLogic());
	}
} routelogic_matcher;

int RouteLogic::command(int argc, const char*const* argv)
{
	if (argc == 2) {
		if (strcmp(argv[1], "compute-routes") == 0) {
			compute_routes();
			return (TCL_OK);
		}
	} else if (argc == 3) {
		if (strcmp(argv[1], "insert") == 0) {
			Node* p = (Node*)TclObject::lookup(argv[2]);
			p->next_ = nodes_;
			nodes_ = p;
			return (TCL_OK);
		}
	}
	return (TclObject::command(argc, argv));
}

void RouteLogic::compute_routes()
{
	Link* link;

	Node* np;
	int n = 0;
	for (np = nodes_; np != 0; np = np->next_) 
		++n;

	n += 1;/* number from 1 */
	int* hopcnt = new int[n];
	int* parent = new int[n];
	int* adj = new int[n * n];
	int* route = new int[n * n];
#define ADJ(i, j) adj[n * (i) + (j)]
#define ROUTE(i, j) route[n * (i) + (j)]

	memset((char *)route, 0, n * n * sizeof(*route));
	int v;
	for (v = 0; v < n * n; ++v)
		adj[v] = INFINITY;

	for (np = nodes_; np != 0; np = np->next_) {
		int v = np->addr() + 1;
		for (Link* link = np->links(); link != 0; link = link->next_)
			ADJ(v, link->neighbor()->addr() + 1) = 1;
	}
	/* do for all the sources */
	int k;
	for (k = 1; k < n; ++k) {
		for (v = 0; v < n; v++)
			parent[v] = v;

		/* set the route for all neighbours first */
		for (v = 1; v < n; ++v) {
			if (parent[v] != k) {
				hopcnt[v] = ADJ(v, k);
				if (hopcnt[v] != INFINITY)
					ROUTE(k, v) = v;
			}
		}
		for (v = 1; v < n; ++v) {
			/*
			 * w is the node that is the nearest to the subtree
			 * that has been routed
			 */
			int o = 0;
			/* XXX */
			hopcnt[0] = INFINITY;
			int w;
			for (w = 1; w < n; w++)
				if (parent[w] != k && hopcnt[w] < hopcnt[o])
					o = w;
			parent[o] = k;
			/*
			 * update distance counts for the nodes that are
			 * adjacent to o
			 */
			if (o == 0)
				continue;
			for (w = 1; w < n; w++) {
				if (parent[w] != k &&
				    hopcnt[o] + ADJ(o, w) < hopcnt[w]) {
					ROUTE(k, w) = ROUTE(k, o);
					hopcnt[w] = hopcnt[o] + 1;
				}
			}
		}
	}
	for (np = nodes_; np != 0; np = np->next_) {
		int v = np->addr() + 1;
		for (int w = 1; w < n; ++w) {
			int o = ROUTE(v, w);
			if (o == 0)
				continue;
			/* Find edge (v, o) */
			for (link = np->links(); link != 0;
			     link = link->next_)
				if (link->neighbor()->addr() + 1 == o)
					break;
			if (link == 0) {
				fprintf(stderr, "ns: routing alg broken\n");
				abort();
			}
			np->setroute(w - 1, link);
		}
	}
	delete hopcnt;
	delete adj;
	delete parent;
	delete route;
}
