/*
 * Copyright (c) 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Lawrence Berkeley Laboratory,
 * Berkeley, CA.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
/*
 * PSC Changes Copyright (c) 1996 Pittsburgh Supercomputing Center,
 *                                Jamshid Mahdavi, and Matt Mathis
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the modifications to this
 * software were developed by the Pittsburgh Supercomputing Center.
 * The name of the Center may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE
 * IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. 
 *
 */

#include <stdio.h>
#include <sys/types.h>

#include "tcp.h"
#include "scoreboard.h"
#include "random.h"
#include "trace.h"

//#define DEBUGSACK1A

#define TRUE    1
#define FALSE   0
#define RECOVER_DUPACK  1
#define RECOVER_TIMEOUT 2
#define RECOVER_QUENCH  3

class FackTcpAgent : public TcpAgent {
 public:
	FackTcpAgent();
	virtual int window();
	virtual void recv(Packet *pkt);
	virtual void timeout(int tno);
	virtual void closecwnd(int how);
	virtual void opencwnd();
	void oldack (Packet* pkt);
	int maxsack (Packet* pkt);
	void plot();
	void send(int force, int reason, int maxburst);
 protected:
	int timeout_;        /* boolean: sent pkt from timeout? */
	int fack_;
	double wintrim_;		/*  Used during rampdown  */
	double wintrimmult_;		/*  Used during rampdown  */
	int ss_div4_;
	int rampdown_;
	int fastrecov_;
	int retran_data_;
	ScoreBoard scb_;
};

static class FackTcpMatcher : public Matcher {
public:
	FackTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-fack") == 0)
			return (new FackTcpAgent());
		return (0);
	}
} matcher_fack;

int FackTcpAgent::window()
{
        int win;

	win = int((cwnd_ < wnd_ ? cwnd_ : wnd_) + wintrim_);
	
        return (win);
}

FackTcpAgent::FackTcpAgent() : rampdown_(0), ss_div4_(0), retran_data_(0), fack_(-1), wintrim_(0), wintrimmult_(.5)
{
	link_int("ns_facktcp", "ss-div4", &ss_div4_, 0);
	link_int("ns_facktcp", "rampdown", &rampdown_, 0);
}

/*
 * Process a dupack.
 */
void FackTcpAgent::oldack(Packet* pkt)
{
	last_ack_ = pkt->seqno_;
	highest_ack_ = last_ack_;
	if (seqno_ < last_ack_ + 1)
		seqno_ = last_ack_ + 1;
	newtimer(pkt);
	if (rtt_active_ && pkt->seqno_ >= rtt_seq_) {
		rtt_active_ = 0;
		t_backoff_ = 1;
	}
	/* with timestamp option */
	double tao = Scheduler::instance().clock() - pkt->bd_.tcp_.ts_;
	rtt_update(tao);
	/* update average window */
	awnd_ *= 1.0 - wnd_th_;
	awnd_ += wnd_th_ * cwnd_;
}

int FackTcpAgent::maxsack(Packet *pkt) 
{
	int maxsack=-1, sack_index;
	
	for (sack_index=0; sack_index < pkt->bd_.tcp_.sa_length_; sack_index++) {
		if (pkt->bd_.tcp_.sa_right_[sack_index] > maxsack)
			maxsack = pkt->bd_.tcp_.sa_right_[sack_index];
	}
	return (maxsack-1);
}
	
void FackTcpAgent::recv(Packet *pkt)
{
	int xmit_seqno, ms;
	int dontSend = 0;

	if (pkt->type_ != PT_ACK) {
		fprintf(stderr,
			"ns: confiuration error: tcp received non-ack\n");
		exit(1);
	}

	if (pkt->flags_ & PF_ECN)
		quench(1);

#ifdef DEBUGSACK1A
			printf("\npkt: seqno_=%d sack_length_=%d {", pkt->seqno_, pkt->bd_.tcp_.sa_length_);
			for (int i=0; i<pkt->bd_.tcp_.sa_length_; i++) {
				printf ("<%d %d> ", pkt->bd_.tcp_.sa_left_[i], pkt->bd_.tcp_.sa_right_[i]);
			}
			printf ("} timeout=%d\n", timeout_);
#endif

        if (!fastrecov_) {
                /* normal... not fast recovery */
                if ((int)pkt->seqno_ > last_ack_ && 
		    pkt->bd_.tcp_.sa_length_ == 0) {
                        /*
                         * regular ACK not in fast recovery... normal
                         */
		        newack(pkt);
			fack_ = last_ack_;
                        opencwnd();
                        timeout_ = FALSE;
			scb_.ClearScoreBoard();
			retran_data_ = 0;
			wintrim_ = 0;
                } 
		else {
			retran_data_ -= scb_.UpdateScoreBoard (highest_ack_, pkt);

			oldack(pkt);
			ms = maxsack(pkt);
			if (ms > fack_) fack_ = ms;

			if (fack_ >= seqno_) 
				seqno_ = fack_ + 1;
			dupacks_ = (fack_ - last_ack_) - 1;

                        /*
                         * a duplicate ACK
                         */
                        if (dupacks_ >= NUMDUPACKS) {
                                /*
                                 * Assume we dropped just one packet.
                                 * Retransmit last ack + 1
                                 * and try to resume the sequence.
                                 */
				recover_ = seqno_;
				recover_cause_ = RECOVER_DUPACK;
				closecwnd(1);
				if (rampdown_) {
					wintrim_ = (seqno_ - fack_ - 1) * wintrimmult_;
				}
				reset_rtx_timer(1);
				fastrecov_ = TRUE;
				scb_.MarkRetran (last_ack_+1, seqno_);
				retran_data_++;
				output(last_ack_ + 1, REASON_DUPACK);
			}
                }
                if (dupacks_ == 0)
                        send(FALSE, REASON_NORMAL, maxburst_);
        } else {
                /* we are in fast recovery */
		oldack(pkt);
		ms = maxsack(pkt);
		if (ms > fack_) {
			if (rampdown_) {
				wintrim_ -= ((float)ms - (float)fack_) * wintrimmult_; 
				if (wintrim_<0) wintrim_ = 0;
			}
			fack_ = ms;
		}

		if (fack_ >= seqno_) 
			seqno_ = fack_ + 1;

		retran_data_ -= scb_.UpdateScoreBoard (highest_ack_, pkt);
		timeout_ |= scb_.CheckSndNxt (pkt);

		opencwnd();

		if (retran_data_ < 0) {
			printf ("Error, retran_data_ < 0");
		}

                if ((int)pkt->bd_.tcp_.sa_length_==0 && (last_ack_ >= recover_)) {
                        /* No SACK blocks indicates fast recovery is over */
			fastrecov_ = FALSE;
                        timeout_ = FALSE;
			scb_.ClearScoreBoard();
			retran_data_ = 0;
			wintrim_ = 0;
			dupacks_=0;
		}

                send(FALSE, REASON_NORMAL, maxburst_);
        }

	Packet::free(pkt);
	if (trace_)
		plot();
}

void FackTcpAgent::timeout(int tno)
{
	if (tno == TCP_TIMER_RTX) {
		timeout_ = FALSE;
		if (highest_ack_ > last_ack_)
			last_ack_ = highest_ack_;
#ifdef DEBUGSACK1A
		printf ("timeout. highest_ack: %d seqno: %d\n", 
		  highest_ack_, seqno_);
#endif
		retran_data_ = 0;
		recover_cause_ = 2;
		closecwnd(0);			/*  Need this??? */
		scb_.ClearScoreBoard();
		reset_rtx_timer(1);
		fack_ = last_ack_;
		seqno_ = last_ack_ + 1;
		send(0, REASON_TIMEOUT);
	} else {
		TcpAgent::timeout(tno);
	}
}

void FackTcpAgent::send(int force, int reason, int maxburst)
{
        register int pktno, found, nextpktno, npacket = 0;
        int win = window();
	int xmit_seqno;

        found = 1;
        /*
         * as long as the pipe is open and there is app data to send...
         */
	while ((seqno_ <= fack_ + win - retran_data_) && (!timeout_)) {
                if (overhead_ == 0 || force) {
                        found = 0;

			xmit_seqno = scb_.GetNextRetran ();
#ifdef DEBUGSACK1A
			printf("highest_ack: %d xmit_seqno: %d timeout: %d seqno: %d fack: %d win: %d retran_data: %d\n", 
			highest_ack_, xmit_seqno, timeout_, seqno_, fack_, win, retran_data_);
#endif
			if (xmit_seqno == -1) { 
				found = 1;
				xmit_seqno = seqno_++;
#ifdef DEBUGSACK1A
				printf("sending %d fastrecovery: %d win %d\n",
				  xmit_seqno, fastrecov_, win);
#endif
			}
			else {
				found = 1;
				scb_.MarkRetran (xmit_seqno, seqno_);
				retran_data_++;
				win = window();
			}
			if (found) {
                        	output(xmit_seqno, reason);
                                if (seqno_ <= xmit_seqno) {
					printf ("Hit a strange case 2.\n");
                                        seqno_ = xmit_seqno + 1;
				}
				npacket++;
			}
                } else if (!pending_[TCP_TIMER_DELSND]) {
                        /*
                         * Set a delayed send timeout.
                         */
                        sched(Random::uniform(overhead_), TCP_TIMER_DELSND);
                        return;
                }
                if (maxburst && npacket == maxburst)
                        break;
        } /* while */
}

void FackTcpAgent::closecwnd(int how)
{   
	if (how == 1) {
		if ((ss_div4_ == 1) && (cwnd_ <= ssthresh_ + .5)) {
			cwnd_ /= 2;
			wintrimmult_ = .75;
		} else {
			wintrimmult_ = .5;
		}
	}
	TcpAgent::closecwnd(how);
}

/*
 * open up the congestion window -- Hack hack hack
 */
void FackTcpAgent::opencwnd()
{
	if (cwnd_ < ssthresh_)
		/* exponential */
		cwnd_ += 1;
	else {
		/* linear */
		double f;
		switch (wnd_option_) {
		case 0:
			if (++count_ >= cwnd_) {
				count_ = 0;
				++cwnd_;
			}
			break;

		case 1:
			/* This is the standard algorithm. */
			if (!fastrecov_)
				cwnd_ += 1 / cwnd_;
			break;

		case 2:
			/* These are window increase algorithms
			 * for experimental purposes only. */
			f = (t_srtt_ >> 3) * tcp_tick_;
			f *= f;
			f *= wnd_const_;
			f += fcnt_;
			if (f > cwnd_) {
				fcnt_ = 0;
				++cwnd_;
			} else
				fcnt_ = f;
			break;

		case 3:
			f = awnd_;
			f *= f;
			f *= wnd_const_;
			f += fcnt_;
			if (f > cwnd_) {
				fcnt_ = 0;
				++cwnd_;
			} else
				fcnt_ = f;
			break;

                case 4:
                        f = awnd_;
                        f *= wnd_const_;
                        f += fcnt_;
                        if (f > cwnd_) {
                                fcnt_ = 0;
                                ++cwnd_;
                        } else
                                fcnt_ = f;
                        break;
		case 5:
                        f = (t_srtt_ >> 3) * tcp_tick_;
                        f *= wnd_const_;
                        f += fcnt_;
                        if (f > cwnd_) {
                                fcnt_ = 0;
                                ++cwnd_;
                        } else
                                fcnt_ = f;
                        break;

		default:
#ifdef notdef
			/*XXX*/
			error("illegal window option %d", wnd_option_);
#endif
			abort();
		}
	}
	// if maxcwnd_ is set (nonzero), make it the cwnd limit
	if (maxcwnd_ && (int (cwnd_) > maxcwnd_))
	  cwnd_ = maxcwnd_;
}

void FackTcpAgent::plot()
{
	double t = Scheduler::instance().clock();
	sprintf(trace_->buffer(), "t %g %d rtt %g\n", 
		t, class_, t_rtt_ * tcp_tick_);
	trace_->dump();
	sprintf(trace_->buffer(), "t %g %d dev %g\n", 
		t, class_, t_rttvar_ * tcp_tick_);
	trace_->dump();
	sprintf(trace_->buffer(), "t %g %d win %f\n", t, class_, cwnd_ + wintrim_);
	trace_->dump();
	sprintf(trace_->buffer(), "t %g %d bck %d\n", t, class_, t_backoff_);
	trace_->dump();
}
